"""
Meetings app

"""
import datetime
from django.db import models
from django.db.models import signals as django_signals
from django.contrib.auth.models import User
from django.contrib.contenttypes.models import ContentType
from django.utils.translation import ugettext_lazy as _

from utils import reports as utils_reports
from utils.versioning import models as versioning_models

class Meeting(versioning_models.VersionedModel):
    date = models.DateField(verbose_name = _('Date'))
    subject = models.CharField(verbose_name=_('Subject'), max_length=100)
    description = models.TextField(verbose_name = _('Description'), blank=True, null=True)
    participants = versioning_models.VersionedManyToManyField(User,
        through='Meeting_User', target_field_name='user', null=True,
        blank=True)


    class Meta:
        verbose_name = _('Meeting')
        verbose_name_plural = _('Meetings')
        ordering = ['-date']
        permissions = (
            ("display_meetings", "May display meetings information"),
            ("edit_meetings", "May edit meetings information"),
            ("participate_meetings", "May participate in meetings"),
        )

    def __unicode__(self):
        from django.conf import settings
        return u'%s - %s' % (datetime.date.strftime(self.date, settings.DATE_OUTPUT_FORMAT), self.subject)

    # urls

    @models.permalink
    def get_absolute_url(self):
        return ('meetings_meeting', (self.v_id,))

    @models.permalink
    def get_edit_url(self):
        return ('meetings_meeting_edit', (self.v_id,))

    @models.permalink
    def get_delete_url(self):
        return ('meetings_meeting_delete', (self.v_id,))

    @models.permalink
    def get_report_url(self):
        return ('meetings_meeting_report', (self.v_id,))

    @models.permalink
    def get_compare_url(self, version):
        return ('meetings_meeting_compare', (self.v_id, version.id))


    # permissions
    @classmethod
    def may_display(cls, user):
        return user.has_perm('meetings.display_meetings')

    @classmethod
    def may_add(cls, user):
        return user.has_perm('meetings.edit_meetings')

    def may_change(self, user):
        if user.has_perm('meetings.edit_meetings'):
            return not self.version
        return False

    def may_delete(self, user):
        return self.may_change(user)

    @classmethod
    def get_listreport_formfields(cls, user):
        from django import forms
        res = [{'name': 'meetings',
            'field': forms.BooleanField(required=False, label=_('Meeting sheets'),),
            'children': cls.get_report_formfields(user)}]
        return res

    @classmethod
    def get_listreport_css(cls, language=u'en', choices={}, **kwargs):
        return ['meetings/meetings_report.css']

    @classmethod
    def get_report_formfields(cls, user):
        return []

    def get_report_body(self, user, language=u'en', choices={}, level=1):
        from meetings import forms as app_forms
        form = app_forms.MeetingSheet(instance=self, omit_empty=True, report=True)
        context = {'object': self, 'form': form, 'title': u'%s: %s' % (_('Meeting'), self), 'level': level}
        html = utils_reports.make_pdf_body(context, language=language)
        return [html,]

    @classmethod
    def get_report_css(cls, language=u'en', choices={}):
        css = super(Meeting, cls).get_report_css(language=language, choices=choices)
        return css

    def get_report(self, user, language=u'en', choices={}):
        body_strings = self.get_report_body(user, language, choices)
        css_files = self.__class__.get_report_css(language, choices)
        return utils_reports.make_pdf(body_strings, css_files, language)

# Explicit m2m model
class Meeting_User(versioning_models.VersionedModel):
    meeting = versioning_models.VersionedForeignKey(Meeting, verbose_name = _('Meeting'))
    user = models.ForeignKey(User, verbose_name = _('Participant'))

    class Meta:
        verbose_name = _('Participant for meeting')
        verbose_name_plural = _('Participants for meeting')

    def __unicode__(self):
        return u'%s - %s' % (self.user, self.meeting)

    # custom methods

    @classmethod
    def collect_items(cls, version):
        modified = list(cls.objects.filter(version=version).values_list('meeting_id', flat=True).distinct())
        if modified:
            items = Meeting.objects.filter(v_id__in=modified).get_versionedobjects(version=version)
            return list(('Participants for meeting', item, None) for item in items)
        return []


def make_meetings_version(sender, instance, objects, **kwargs):
    if not objects:
        # only version everything when objects are not specifically included
        if instance.published:
            filterkwarg={'published_dirty': True, 'version__isnull': True}
        else:
            filterkwarg={'dirty': True, 'version__isnull': True}
        for item in Meeting.objects.filter(**filterkwarg):
            item.make_version(instance)
        for item in Meeting_User.objects.filter(**filterkwarg):
            item.make_version(instance)

versioning_models.post_version.connect(make_meetings_version, sender=versioning_models.Version)

def restore_meetings_version(sender, instance, objects, **kwargs):
    if not objects:
        for item in Meeting.objects.get_newerobjects(instance):
            item.restore_version(instance)
        for item in Meeting_User.objects.get_newerobjects(instance):
            item.restore_version(instance)

versioning_models.restore_version.connect(restore_meetings_version, sender=versioning_models.Version)

from utils.search.models import searches, Search

class Meetings_MeetingSearch(Search):
    permission = 'meetings.display_meetings'
    fields = ['description',]
    model = Meeting

searches.register(Meetings_MeetingSearch)


