from django import forms
from django.contrib.auth.models import User, Group, Permission
from django.contrib.contenttypes.models import ContentType
from django.db.models import Q
from django.utils.translation import ugettext_lazy as _

from installations.models import SectionTreeElement
from utils import fields as utils_fields
from utils import lists as utils_lists
from utils import widgets as utils_widgets
from utils import sheets
from tasks.models import Task
from utils.versioning import versionedforms


class TaskForm(versionedforms.VersionedForm):
    user = utils_fields.UserModelChoiceField(User.objects.none(),
        label=_('Responsable'),
        required=False,
    )
    class Meta:
        model = Task
        fields = ('task', 'description', 'result', 'duedate', 'completiondate', 'user')
        attrs = {
            'task': {'size': 70},
            'description': {'rows': 10, 'cols': 60},
            'result': {'rows': 10, 'cols': 60},
        }
        widgets = {
            'description': utils_widgets.MarkItUpWidget(),
            'result': utils_widgets.MarkItUpWidget()
        }
    def __init__(self, *args, **kwargs):
        super(TaskForm, self).__init__(*args, **kwargs)
        _gettaskpermission = Permission.objects.get_by_natural_key('get_tasks', 'tasks', 'task')
        self.fields['user'].queryset = User.objects.filter(
            Q(user_permissions=_gettaskpermission)|Q(groups__permissions=_gettaskpermission)
            ).distinct()

class TaskSheet(versionedforms.VersionedSheet):
    content_object = versionedforms.VersionedForeignKeyField(
            data = 'content_object',
            name = 'content_object',
            label = _('Related object'),
            add_link = True
        )
    description = utils_fields.MarkdownField(data = 'description', name='description', label=_('Description'))
    result = utils_fields.MarkdownField(data = 'result', name='result', label=_('Result'))
    user = utils_fields.UserSheetForeignKeyField(data='user', name='user', label=_('Responsable'))
    class Meta:
        model = Task
        fields = ('content_object', 'task', 'description', 'result', 'duedate', 'completiondate', 'user')

    def __init__(self, instance, *args, **kwargs):
        super(TaskSheet, self).__init__(instance, *args, **kwargs)
        if instance:
            if instance.content_object:
                self.fields['content_object'].label = instance.content_object.__class__._meta.verbose_name

class TaskFilterForm(utils_lists.FilterForm):
    """Custom filterform for the tasks list"""
    task = forms.CharField(max_length=50, required=False, label=_('Task'))
    user = utils_fields.UserModelChoiceField(queryset = User.objects.none(), required=False,
        label=_('Responsable'))
    installation = forms.ModelChoiceField(queryset=SectionTreeElement.objects.filter(parent__isnull=True).order_by('tree_id'), required=False,
        label=_('Installation'))
    complete = forms.ChoiceField(choices=utils_lists.get_filter_choices([(1, _('Yes')), (2, _('No')), (3, _('Incomplete, late')), (4, _('Completed on time')), (5, _('Completed late'))]), required=False, label = _('Complete'))

    class Meta:
        attrs = {
            'task': {'size': 40},
            }

    def __init__(self, version=None, *args, **kwargs):
        super(TaskFilterForm, self).__init__(*args, **kwargs)
        _gettaskpermission = Permission.objects.get_by_natural_key('get_tasks', 'tasks', 'task')
        self.fields['user'].queryset = User.objects.filter(
            Q(user_permissions=_gettaskpermission)|Q(groups__permissions=_gettaskpermission)
            ).distinct().order_by('last_name', 'first_name')
        if version:
            self.fields['installation'].queryset = SectionTreeElement.objects.filter(parent__isnull=True).order_by('tree_id').get_versionedobjects(version)
