from django.conf import settings
from django.contrib.auth.models import User
from django.core import exceptions
from django.db import connection
from django.test import TestCase

from mptt.utils import tree_item_iterator

import time

from utils import tests as utils_tests

from utils.versioning.models import Version, WorkVersion
from utils.versioning.versionedforms import VersionedForm

from utils.versioning.versioning_test import models as app_models

def print_tree(items):
    res =[]
    prefix = ''
    for element,structure in tree_item_iterator(items):
        if structure['new_level']:
            prefix = '-%s' % prefix
        res.append(u'%s%s' % (prefix, element))
        x = len(structure['closed_levels'])
        if x:
            prefix = prefix[:-x]
    return u'|'.join(res)

class TestVTestCase(TestCase):
    fixtures = ['dummyusers',]

class Simple(TestVTestCase):
    def test_versioning(self):
        user = User.objects.get(pk=1)

        self.assertEquals(app_models.Building.objects.count(), 0)
        self.assertEquals(app_models.Building.objects.next_v_id(), 1)

        b1 = app_models.Building(name='b1')
        b1.save(user)

        b2 = app_models.Building(name='b2')
        b2.save(user)

        self.assertEquals(app_models.Building.objects.count(), 2)
        self.assertEquals(app_models.Building.objects.dirty().count(), 2)

        v1 = Version(name='v1')
        v1.save(user)

        self.assertEquals(app_models.Building.objects.count(), 4)
        self.assertEquals(app_models.Building.objects.dirty().count(), 0)

        self.assertEquals(app_models.Building.objects.next_v_id(), 3)

        b1 = app_models.Building.objects.get_versionedobject(v_id = 1, version=WorkVersion)
        b1.name = 'b1a'
        b1.save(user)
        self.assertEquals(b1.v_id, 1)

        self.assertEquals(app_models.Building.objects.count(), 4)
        self.assertEquals(app_models.Building.objects.dirty().count(), 1)

        # wait a bit to distinguish the datetimes on the versions
        time.sleep(0.01)
        v2 = Version(name='v2')
        v2.save(user)

        self.assertEquals(app_models.Building.objects.count(), 5)
        self.assertEquals(app_models.Building.objects.dirty().count(), 0)

        # let's delete b2
        b2 = app_models.Building.objects.get_versionedobject(v_id=2, version=WorkVersion)
        b2.delete(user)

        self.assertEquals(app_models.Building.objects.count(), 5)
        self.assertEquals(app_models.Building.objects.dirty().count(), 1)
        self.assertEquals(app_models.Building.objects.filter(deleted=True).count(), 1)

        # wait a bit to distinguish the datetimes on the versions
        time.sleep(0.01)
        v3 = Version(name='v3')
        v3.save(user)

        self.assertEquals(app_models.Building.objects.count(), 5)
        self.assertEquals(app_models.Building.objects.dirty().count(), 0)
        self.assertEquals(app_models.Building.objects.filter(deleted=True).count(), 1)

        # let's add a app_models.Building and delete it again
        b3 = app_models.Building(name='b3')
        b3.save(user)
        self.assertEquals(b3.v_id, 3)
        self.assertEquals(app_models.Building.objects.count(), 6)
        self.assertEquals(app_models.Building.objects.dirty().count(), 1)
        b3 = app_models.Building.objects.get_versionedobject(v_id=3, version=WorkVersion)
        b3.delete(user)
        # b3 should dissappear entirely
        self.assertEquals(app_models.Building.objects.count(), 5)
        self.assertEquals(app_models.Building.objects.dirty().count(), 0)

        # wait a bit to distinguish the datetimes on the versions
        time.sleep(0.01)
        v4 = Version(name='v4')
        v4.save(user)

        self.assertEquals(app_models.Building.objects.count(), 5)
        self.assertEquals(app_models.Building.objects.dirty().count(), 0)
        self.assertEquals(app_models.Building.objects.filter(deleted=True).count(), 1)

        # now let's check the manager

        qs = app_models.Building.objects.all().get_versionedobjects(WorkVersion).order_by('name')
        # b1a
        self.assertEquals(qs.count(), 1)
        self.assertEquals(qs[0].name, 'b1a')

        qs = app_models.Building.objects.all().get_versionedobjects(v1).order_by('name')
        # original b1 and b2
        self.assertEquals(qs.count(), 2)
        self.assertEquals(qs[0].name, 'b1')
        self.assertEquals(qs[1].name, 'b2')

        qs = app_models.Building.objects.all().get_versionedobjects(v2).order_by('name')
        # b1a and b2
        self.assertEquals(qs.count(), 2)
        self.assertEquals(qs[0].name, 'b1a')
        self.assertEquals(qs[1].name, 'b2')

        qs = app_models.Building.objects.all().get_versionedobjects(v3).order_by('name')
        # b1a
        self.assertEquals(qs.count(), 1)
        self.assertEquals(qs[0].name, 'b1a')

        qs = app_models.Building.objects.all().get_versionedobjects(v4).order_by('name')
        # B and C
        # b1a
        self.assertEquals(qs.count(), 1)
        self.assertEquals(qs[0].name, 'b1a')

        # check collect_items
        self.assertEquals(len(v1.collect_items(user)), 2) # 2 buildings
        self.assertEquals(len(v2.collect_items(user)), 1)
        self.assertEquals(len(v3.collect_items(user)), 1)
        self.assertEquals(len(v4.collect_items(user)), 0)

class ForeignKeys(TestVTestCase):
    def test_versioningfks(self):
        user = User.objects.get(pk=1)

        self.assertEquals(app_models.Building.objects.count(), 0)
        self.assertEquals(app_models.Appartment.objects.count(), 0)
        self.assertEquals(app_models.Building.objects.next_v_id(), 1)
        self.assertEquals(app_models.Appartment.objects.next_v_id(), 1)

        b1 = app_models.Building(name='b1')
        b1.save(user)

        a1 = app_models.Appartment(name='a1', building=b1)
        a1.save(user)

        v1 = Version(name='v1')
        v1.save(user)

        self.assertEquals(app_models.Building.objects.count(), 2)
        self.assertEquals(app_models.Building.objects.dirty().count(), 0)
        self.assertEquals(app_models.Building.objects.next_v_id(), 2)

        self.assertEquals(app_models.Appartment.objects.count(), 2)
        self.assertEquals(app_models.Appartment.objects.dirty().count(), 0)
        self.assertEquals(app_models.Appartment.objects.next_v_id(), 2)

        a1 = app_models.Appartment.objects.get_versionedobject(v_id=1, version=WorkVersion)
        b1 = app_models.Building.objects.get_versionedobject(v_id=1, version=WorkVersion)
        self.assertEquals(a1.building, b1)

        a1a = app_models.Appartment.objects.get_versionedobject(v_id=1, version=v1)
        b1a = app_models.Building.objects.get_versionedobject(v_id=1, version=v1)
        self.assertEquals(a1a.building, b1a)

        # modify building 1
        b1 = app_models.Building.objects.get_versionedobject(v_id=1, version=WorkVersion)
        b1.name = 'b1b'
        b1.save(user)
        self.assertEquals(app_models.Building.objects.count(), 2)
        self.assertEquals(app_models.Building.objects.dirty().count(), 1)

        # wait a bit to distinguish the datetimes on the versions
        time.sleep(0.01)
        v2 = Version(name='v2')
        v2.save(user)

        self.assertEquals(app_models.Building.objects.count(), 3)
        self.assertEquals(app_models.Building.objects.dirty().count(), 0)

        #let 's look at the appartment
        a = app_models.Appartment.objects.get_versionedobject(v_id=1, version=WorkVersion)
        self.assertEquals(a.building.name, 'b1b')

        a = app_models.Appartment.objects.get_versionedobject(v_id=1, version=v1)
        self.assertEquals(a.building.name, 'b1')


class ForeignKeysReversed(TestVTestCase):
    def test_versioningreversefks(self):
        user = User.objects.get(pk=1)

        self.assertEquals(app_models.Building.objects.count(), 0)
        self.assertEquals(app_models.Appartment.objects.count(), 0)
        self.assertEquals(app_models.Building.objects.next_v_id(), 1)
        self.assertEquals(app_models.Appartment.objects.next_v_id(), 1)

        b1 = app_models.Building(name='b1')
        b1.save(user)

        a1 = app_models.Appartment(name='a1', building=b1)
        a1.save(user)

        v1 = Version(name='v1')
        v1.save(user)

        # modify appartment 1
        a1 = app_models.Appartment.objects.get_versionedobject(v_id=1, version=WorkVersion)
        a1.name = 'a1b'
        a1.save(user)
        self.assertEquals(app_models.Appartment.objects.count(), 2)
        self.assertEquals(app_models.Appartment.objects.dirty().count(), 1)

        # wait a bit to distinguish the datetimes on the versions
        time.sleep(0.01)
        v2 = Version(name='v2')
        v2.save(user)

        self.assertEquals(app_models.Appartment.objects.count(), 3)
        self.assertEquals(app_models.Appartment.objects.dirty().count(), 0)

        b1 = app_models.Building.objects.get_versionedobject(v_id=1, version=WorkVersion)

        a = app_models.Appartment.objects.filter(building=b1).get_versionedobjects(version=WorkVersion)
        self.assertEqual(len(a), 1)
        self.assertEquals(a[0].name, 'a1b')

        a = app_models.Appartment.objects.filter(building=b1).get_versionedobjects(version=v2)
        self.assertEqual(len(a), 1)
        self.assertEquals(a[0].name, 'a1b')

        a = app_models.Appartment.objects.filter(building=b1).get_versionedobjects(version=v1)
        self.assertEqual(len(a), 1)
        self.assertEquals(a[0].name, 'a1')

        # test related descriptor
        b1 = app_models.Building.objects.get_versionedobject(v_id=1, version=WorkVersion)
        a = b1.appartment_set.all()
        self.assertEqual(len(a), 1)
        self.assertEquals(a[0].name, 'a1b')

        b1 = app_models.Building.objects.get_versionedobject(v_id=1, version=v2)
        a = b1.appartment_set.all()
        self.assertEqual(len(a), 1)
        self.assertEquals(a[0].name, 'a1b')

        b1 = app_models.Building.objects.get_versionedobject(v_id=1, version=v1)
        a = b1.appartment_set.all()
        self.assertEqual(len(a), 1)
        self.assertEquals(a[0].name, 'a1')

class Tree(TestVTestCase):
    def test_treemodels(self):
        user = User.objects.get(pk=1)

        self.assertEquals(app_models.DiskTreeElement.objects.count(), 0)
        self.assertEquals(app_models.Folder.objects.count(), 0)

        # create a tree element
        folder = app_models.Folder(name=u'Base')
        treeelement = folder.save(user, parent=None, treeelementkwargs={})
        self.assertEquals(folder.tree_display(), u'Base')
        self.assertEquals(folder.get_elementtype_for_tree(), u'event')
        self.assertEquals(folder.get_treeelements().count(), 1)

        # and another
        folder = app_models.Folder(name=u'Child 1')
        folder.save(user, parent=treeelement, treeelementkwargs={})

        # and another
        folder = app_models.Folder(name=u'Child 2')
        folder.save(user, parent=treeelement, treeelementkwargs={})

        self.assertEquals(app_models.DiskTreeElement.objects.count(), 3)
        self.assertEquals(app_models.Folder.objects.count(), 3)

        tree = app_models.DiskTreeElement.tree.get_versionedobjects(version=WorkVersion)
        self.assertEquals(tree.count(), 3)

        # let's test the VGFK:
        el1 = app_models.DiskTreeElement.objects.get_versionedobject(v_id=1, version=WorkVersion)
        self.assertEquals(el1.content_object.name, u'Base')
        self.assertEquals(len(el1.content_object.get_treeelements()), 1)
        self.assertEquals(el1.content_object.get_treeelements()[0], el1)

        v1 = Version(name='v1')
        v1.save(user)

        self.assertEquals(app_models.DiskTreeElement.objects.count(), 6)
        self.assertEquals(app_models.Folder.objects.count(), 6)

        tree = app_models.DiskTreeElement.tree.get_versionedobjects(version=WorkVersion)
        self.assertEquals(tree.count(), 3)
        tree = app_models.DiskTreeElement.tree.get_versionedobjects(version=v1)
        self.assertEquals(tree.count(), 3)

        # test modifying a leaf
        folder = app_models.Folder.objects.get_versionedobject(name='Child 1', version=WorkVersion)
        folder.name = u'Modified child 1'
        folder.save(user)

        # wait a bit to distinguish the datetimes on the versions
        time.sleep(0.01)
        v2 = Version(name='v2')
        v2.save(user)

        self.assertEquals(app_models.DiskTreeElement.objects.count(), 6)
        self.assertEquals(app_models.Folder.objects.count(), 7)

        tree = app_models.DiskTreeElement.tree.get_versionedobjects(version=WorkVersion)
        self.assertEquals(tree.count(), 3)
        self.assertEquals(tree[2].content_object.name, u'Modified child 1')

        tree = app_models.DiskTreeElement.tree.get_versionedobjects(version=v2)
        self.assertEquals(tree.count(), 3)
        self.assertEquals(tree[2].content_object.name, u'Modified child 1')

        tree = app_models.DiskTreeElement.tree.get_versionedobjects(version=v1)
        self.assertEquals(tree.count(), 3)
        self.assertEquals(tree[2].content_object.name, u'Child 1')

        # now test modifying the tree
        # move child 2 underneath child 1
        self.assertEquals(app_models.DiskTreeElement.objects.get_versionedobject(v_id=3, version=WorkVersion).parent.v_id, 1)
        el1 = app_models.DiskTreeElement.objects.get_versionedobject(v_id=2, version=WorkVersion)
        el2 = app_models.DiskTreeElement.objects.get_versionedobject(v_id=3, version=WorkVersion)
        el2.move_to(user, el1)
        self.assertEquals(app_models.DiskTreeElement.objects.get_versionedobject(v_id=3, version=WorkVersion).dirty, True)
        self.assertEquals(app_models.DiskTreeElement.objects.get_versionedobject(v_id=1, version=WorkVersion).dirty, True)

        # wait a bit to distinguish the datetimes on the versions
        time.sleep(0.01)
        v3 = Version(name='v3')
        v3.save(user)

        self.assertEquals(app_models.DiskTreeElement.objects.count(), 9)
        self.assertEquals(app_models.Folder.objects.count(), 7)

        tree = app_models.DiskTreeElement.tree.get_versionedobjects(version=WorkVersion)
        self.assertEquals(tree.count(), 3)
        self.assertEquals(app_models.DiskTreeElement.objects.get_versionedobject(v_id=3, version=WorkVersion).parent.v_id, 2)

        tree = app_models.DiskTreeElement.tree.get_versionedobjects(version=v3)
        self.assertEquals(tree.count(), 3)
        self.assertEquals(app_models.DiskTreeElement.objects.get_versionedobject(v_id=3, version=v3).parent.v_id, 2)

        tree = app_models.DiskTreeElement.tree.get_versionedobjects(version=v2)
        self.assertEquals(tree.count(), 3)
        self.assertEquals(app_models.DiskTreeElement.objects.get_versionedobject(v_id=3, version=v2).parent.v_id, 1)

        # Lets test select_related_versioned
        olddebug = settings.DEBUG
        settings.DEBUG = True
        self.assertEquals(len(connection.queries), 0)
        tree = app_models.DiskTreeElement.tree.get_versionedobjects(version=WorkVersion).select_related_versioned()
        self.assertEquals(len(connection.queries), 3)
        # now access the objects
        self.assertEquals(tree[0].content_object.name, u'Base')
        self.assertEquals(tree[1].content_object.name, u'Modified child 1')
        self.assertEquals(tree[2].content_object.name, u'Child 2')
        self.assertEquals(len(connection.queries), 3)
        settings.DEBUG = olddebug

        # test get_treeelements
        folder = app_models.Folder.objects.get_versionedobject(v_id=1, version=WorkVersion)
        self.assertEquals(folder.get_treeelements()[0].v_id, 1)
        wv_id = folder.get_treeelements()[0].id
        folder = app_models.Folder.objects.get_versionedobject(v_id=1, version=v1)
        self.assertEquals(folder.get_treeelements()[0].v_id, 1)
        v1_id = folder.get_treeelements()[0].id
        self.assertNotEquals(v1_id, wv_id)
        folder = app_models.Folder.objects.get_versionedobject(v_id=1, version=v2)
        v2_id = folder.get_treeelements()[0].id
        self.assertEquals(folder.get_treeelements()[0].v_id, 1)
        self.assertEquals(v1_id, v2_id)
        folder = app_models.Folder.objects.get_versionedobject(v_id=1, version=v3)
        v3_id = folder.get_treeelements()[0].id
        self.assertEquals(folder.get_treeelements()[0].v_id, 1)
        self.assertNotEquals(v3_id, v2_id)

class ManyToMany(TestVTestCase):
    def test_versioningm2m(self):
        user = User.objects.get(pk=1)

        self.assertEquals(app_models.CleaningLady.objects.count(), 0)
        self.assertEquals(app_models.Appartment.objects.count(), 0)
        self.assertEquals(app_models.Appartment_CleaningLady.objects.count(), 0)

        c1 = app_models.CleaningLady(name='c1')
        c1.save(user)

        c2 = app_models.CleaningLady(name='c2')
        c2.save(user)

        b1 = app_models.Building(name='b1')
        b1.save(user)

        a1 = app_models.Appartment(name='a1', building=b1)
        a1.save(user)

        a1_c1 = app_models.Appartment_CleaningLady(appartment=a1, cleaninglady=c1)
        a1_c1.save(user)

        # test descriptor
        self.assertEquals(repr(a1.cleaningladies.all()), u'[<CleaningLady: c1>]')
        # test reversedescriptor
        self.assertEquals(repr(c1.appartment_set.all()), u'[<Appartment: a1>]')

        v1 = Version(name='v1')
        v1.save(user)

        # Change cleaning lady
        a1_c1=app_models.Appartment_CleaningLady.objects.get_versionedobject(appartment=a1, cleaninglady=c1, version=WorkVersion)
        a1_c1.delete(user)

        a1_c2 = app_models.Appartment_CleaningLady(appartment=a1, cleaninglady=c2)
        a1_c2.save(user)

        # test descriptor
        self.assertEquals(repr(a1.cleaningladies.all()), u'[<CleaningLady: c2>]')
        # test reversedescriptor
        self.assertEquals(repr(c2.appartment_set.all()), u'[<Appartment: a1>]')

        a1 = app_models.Appartment.objects.get_versionedobject(name='a1', version=v1)
        self.assertEquals(repr(a1.cleaningladies.all()), u'[<CleaningLady: c1 (v1)>]')
        c1 = app_models.CleaningLady.objects.get_versionedobject(name='c1', version=v1)
        self.assertEquals(repr(c1.appartment_set.all()), u'[<Appartment: a1 (v1)>]')
        c2 = app_models.CleaningLady.objects.get_versionedobject(name='c2', version=v1)
        self.assertEquals(repr(c2.appartment_set.all()), u'[]')

class ManyToManyNormal(TestVTestCase):
    def test_versioningm2mnormal(self):
        user = User.objects.get(pk=1)
        u2 = User.objects.get(pk=2)

        self.assertEquals(app_models.Appartment.objects.count(), 0)
        self.assertEquals(app_models.Appartment_User.objects.count(), 0)

        b1 = app_models.Building(name='b1')
        b1.save(user)

        a1 = app_models.Appartment(name='a1', building=b1)
        a1.save(user)

        a1_u1 = app_models.Appartment_User(appartment=a1, inhabitant=user)
        a1_u1.save(user)

        # test descriptor
        self.assertEquals(repr(a1.inhabitants.all()), u'[<User: u1>]')
        # test reversedescriptor
        self.assertEquals(repr(user.appartment_set.all()), u'[<Appartment: a1>]')

        v1 = Version(name='v1')
        v1.save(user)

        # Change cleaning lady
        a1_u1=app_models.Appartment_User.objects.get_versionedobject(appartment=a1, inhabitant=user, version=WorkVersion)
        a1_u1.delete(user)

        a1_u2 = app_models.Appartment_User(appartment=a1, inhabitant=u2)
        a1_u2.save(user)

        # test descriptor
        a1 = app_models.Appartment.objects.get_versionedobject(name='a1', version=WorkVersion)
        self.assertEquals(repr(a1.inhabitants.all()), u'[<User: u2>]')
        # test reversedescriptor
        self.assertEquals(repr(u2.appartment_set.all()), u'[<Appartment: a1>]')

        a1 = app_models.Appartment.objects.get_versionedobject(name='a1', version=v1)
        self.assertEquals(repr(a1.inhabitants.all()), u'[<User: u1>]')


class ForeignKeyValidation(TestVTestCase):
    def test_validation(self):
        user = User.objects.get(pk=1)

        self.assertEquals(app_models.Building.objects.count(), 0)
        self.assertEquals(app_models.Appartment.objects.count(), 0)
        self.assertEquals(app_models.Building.objects.next_v_id(), 1)
        self.assertEquals(app_models.Appartment.objects.next_v_id(), 1)

        b1 = app_models.Building(name='b1')
        b1.save(user)

        class AppartmentForm(VersionedForm):
            class Meta:
                model = app_models.Appartment

        # using the a valid building v_id
        a1 = AppartmentForm({'name':'a1', 'building': b1.v_id})
        self.assertEquals(a1.is_valid(), True)

        # invalid building v_id
        a2 = AppartmentForm({'name':'a1', 'building': 60})
        self.assertEquals(a2.is_valid(), False)


class TestPublishTestCase(TestCase):
    fixtures = ['dummyusers',]

class PublishSimple(TestPublishTestCase):
    def test_publishsimple(self):
        user = User.objects.get(pk=1)

        self.assertEquals(app_models.Building.objects.count(), 0)
        self.assertEquals(app_models.Building.objects.next_v_id(), 1)

        b1 = app_models.Building(name='b1')
        b1.save(user)

        b2 = app_models.Building(name='b2')
        b2.save(user)

        self.assertEquals(app_models.Building.objects.count(), 2)
        self.assertEquals(app_models.Building.objects.dirty().count(), 2)
        self.assertEquals(app_models.Building.objects.filter(published_dirty=True, version__isnull=True).count(), 2)
        self.assertEquals(app_models.Building.objects.filter(published=True).count(), 0)

        v1 = Version(name='v1')
        v1.save(user)

        self.assertEquals(app_models.Building.objects.count(), 4)
        self.assertEquals(app_models.Building.objects.dirty().count(), 0)
        self.assertEquals(app_models.Building.objects.filter(published_dirty=True, version__isnull=True).count(), 2)
        self.assertEquals(app_models.Building.objects.filter(published=True).count(), 0)

        # wait a bit to distinguish the datetimes on the versions
        time.sleep(0.01)
        # create a public version
        v2 = Version(name='v2', published=True)
        v2.save(user)

        self.assertEquals(app_models.Building.objects.count(), 6)
        self.assertEquals(app_models.Building.objects.dirty().count(), 0)
        self.assertEquals(app_models.Building.objects.filter(published_dirty=True, version__isnull=True).count(), 0)
        self.assertEquals(app_models.Building.objects.filter(published=True).count(), 2)


        b1 = app_models.Building.objects.get_versionedobject(v_id = 1, version=WorkVersion)
        b1.name = 'b1a'
        b1.save(user)
        self.assertEquals(b1.v_id, 1)

        self.assertEquals(app_models.Building.objects.count(), 6)
        self.assertEquals(app_models.Building.objects.dirty().count(), 1)
        self.assertEquals(app_models.Building.objects.filter(published_dirty=True, version__isnull=True).count(), 1)
        self.assertEquals(app_models.Building.objects.filter(published=True).count(), 2)

        time.sleep(0.01)
        v3 = Version(name='v3')
        v3.save(user)

        self.assertEquals(app_models.Building.objects.count(), 7)
        self.assertEquals(app_models.Building.objects.dirty().count(), 0)
        self.assertEquals(app_models.Building.objects.filter(published_dirty=True, version__isnull=True).count(), 1)
        self.assertEquals(app_models.Building.objects.filter(published=True).count(), 2)

        # change b2
        b2 = app_models.Building.objects.get_versionedobject(v_id = 2, version=WorkVersion)
        b2.name = 'b2a'
        b2.save(user)
        self.assertEquals(b2.v_id, 2)

        self.assertEquals(app_models.Building.objects.count(), 7)
        self.assertEquals(app_models.Building.objects.dirty().count(), 1)
        self.assertEquals(app_models.Building.objects.filter(published_dirty=True, version__isnull=True).count(), 2)
        self.assertEquals(app_models.Building.objects.filter(published=True).count(), 2)

        time.sleep(0.01)
        # create a normal version
        v4 = Version(name='v4')
        v4.save(user)

        self.assertEquals(app_models.Building.objects.count(), 8)
        self.assertEquals(app_models.Building.objects.dirty().count(), 0)
        self.assertEquals(app_models.Building.objects.filter(published_dirty=True, version__isnull=True).count(), 2)
        self.assertEquals(app_models.Building.objects.filter(published=True).count(), 2)

        time.sleep(0.01)
        # create public version only of b1
        v5 = Version(name='v5', published=True)
        v5.save(user, propagate=False)
        b1 = app_models.Building.objects.get_versionedobject(v_id = 1, version=WorkVersion)
        b1.make_version(v5)

        self.assertEquals(app_models.Building.objects.count(), 9)
        self.assertEquals(app_models.Building.objects.dirty().count(), 0)
        self.assertEquals(app_models.Building.objects.filter(published_dirty=True, version__isnull=True).count(), 1)
        self.assertEquals(app_models.Building.objects.filter(published=True).count(), 3)

        time.sleep(0.01)
        # now publish all
        v6 = Version(name='v6', published=True)
        v6.save(user)

        self.assertEquals(app_models.Building.objects.count(), 10)
        self.assertEquals(app_models.Building.objects.dirty().count(), 0)
        self.assertEquals(app_models.Building.objects.filter(published_dirty=True, version__isnull=True).count(), 0)
        self.assertEquals(app_models.Building.objects.filter(published=True).count(), 4)

        # now let's check the manager

        qs = app_models.Building.objects.all().get_versionedobjects(WorkVersion).order_by('name')
        # b1a and b2a
        self.assertEquals(qs.count(), 2)
        self.assertEquals(qs[0].name, 'b1a')
        self.assertEquals(qs[1].name, 'b2a')

        qs = app_models.Building.objects.all().get_versionedobjects(v1).order_by('name')
        # original b1 and b2
        self.assertEquals(qs.count(), 2)
        self.assertEquals(qs[0].name, 'b1')
        self.assertEquals(qs[1].name, 'b2')

        qs = app_models.Building.objects.all().get_versionedobjects(v2).order_by('name')
        # This one is published
        # still b1 and b2
        self.assertEquals(qs.count(), 2)
        self.assertEquals(qs[0].name, 'b1')
        self.assertEquals(qs[1].name, 'b2')

        qs = app_models.Building.objects.all().get_versionedobjects(v3).order_by('name')
        # b1a and b2
        self.assertEquals(qs.count(), 2)
        self.assertEquals(qs[0].name, 'b1a')
        self.assertEquals(qs[1].name, 'b2')

        qs = app_models.Building.objects.all().get_versionedobjects(v4).order_by('name')
        # b1a and b2a
        self.assertEquals(qs.count(), 2)
        self.assertEquals(qs[0].name, 'b1a')
        self.assertEquals(qs[1].name, 'b2a')

        qs = app_models.Building.objects.all().get_versionedobjects(v5).order_by('name')
        # This one is published
        # b1a but still b2
        self.assertEquals(qs.count(), 2)
        self.assertEquals(qs[0].name, 'b1a')
        self.assertEquals(qs[1].name, 'b2')

        qs = app_models.Building.objects.all().get_versionedobjects(v6).order_by('name')
        # This one is published
        # b1a and b2a
        self.assertEquals(qs.count(), 2)
        self.assertEquals(qs[0].name, 'b1a')
        self.assertEquals(qs[1].name, 'b2a')

class PublishComplex(TestPublishTestCase):
    def test_publishcomplex(self):
        user = User.objects.get(pk=1)

        self.assertEquals(app_models.Building.objects.count(), 0)
        self.assertEquals(app_models.Appartment.objects.count(), 0)
        self.assertEquals(app_models.CleaningLady.objects.count(), 0)
        self.assertEquals(app_models.Appartment_CleaningLady.objects.count(), 0)

        c1 = app_models.CleaningLady(name='c1')
        c1.save(user)

        c2 = app_models.CleaningLady(name='c2')
        c2.save(user)

        b1 = app_models.Building(name='b1')
        b1.save(user)

        a1 = app_models.Appartment(name='a1', building=b1)
        a1.save(user)

        a1_c1 = app_models.Appartment_CleaningLady(appartment=a1, cleaninglady=c1)
        a1_c1.save(user)

        self.assertEqual(a1.may_publish(user), False)

        # make a public version of everything
        v1 = Version(name='v1', published=True)
        v1.save(user)

        self.assertEquals(app_models.Building.objects.count(), 2)
        self.assertEquals(app_models.Building.objects.dirty().count(), 0)
        self.assertEquals(app_models.Building.objects.filter(published_dirty=True, version__isnull=True).count(), 0)
        self.assertEquals(app_models.Building.objects.filter(published=True).count(), 1)
        self.assertEquals(app_models.Appartment.objects.count(), 2)
        self.assertEquals(app_models.Appartment.objects.dirty().count(), 0)
        self.assertEquals(app_models.Appartment.objects.filter(published_dirty=True, version__isnull=True).count(), 0)
        self.assertEquals(app_models.Appartment.objects.filter(published=True).count(), 1)
        self.assertEquals(app_models.CleaningLady.objects.count(), 4)
        self.assertEquals(app_models.CleaningLady.objects.dirty().count(), 0)
        self.assertEquals(app_models.CleaningLady.objects.filter(published_dirty=True, version__isnull=True).count(), 0)
        self.assertEquals(app_models.CleaningLady.objects.filter(published=True).count(), 2)
        self.assertEquals(app_models.Appartment_CleaningLady.objects.count(), 2)
        self.assertEquals(app_models.Appartment_CleaningLady.objects.dirty().count(), 0)
        self.assertEquals(app_models.Appartment_CleaningLady.objects.filter(published_dirty=True, version__isnull=True).count(), 0)
        self.assertEquals(app_models.Appartment_CleaningLady.objects.filter(published=True).count(), 1)

        # Change cleaning lady
        a1_c1=app_models.Appartment_CleaningLady.objects.get_versionedobject(appartment=a1, cleaninglady=c1, version=WorkVersion)
        a1_c1.delete(user)

        a1_c2 = app_models.Appartment_CleaningLady(appartment=a1, cleaninglady=c2)
        a1_c2.save(user)

        self.assertEquals(app_models.Appartment_CleaningLady.objects.dirty().count(), 2)
        self.assertEquals(app_models.Appartment_CleaningLady.objects.filter(published_dirty=True, version__isnull=True).count(), 2)

        # Change building 1
        b1 = app_models.Building.objects.get_versionedobject(v_id = 1, version=WorkVersion)
        b1.name = 'b1a'
        b1.save(user)

        self.assertEquals(app_models.Building.objects.dirty().count(), 1)
        self.assertEquals(app_models.Building.objects.filter(published_dirty=True, version__isnull=True).count(), 1)

        # Now let's publish appartment a1, but not b1
        a1 = app_models.Appartment.objects.get_versionedobject(v_id = 1, version=WorkVersion)

        self.assertEquals(a1.may_publish(user), True)

        v2 = Version(name='v2', published=True)
        v2.save(user, propagate=False)
        a1.publish(user, v2)

        # b1 still dirty and published_dirty
        self.assertEquals(app_models.Building.objects.count(), 2)
        self.assertEquals(app_models.Building.objects.dirty().count(), 1)
        self.assertEquals(app_models.Building.objects.filter(published_dirty=True, version__isnull=True).count(), 1)
        self.assertEquals(app_models.Building.objects.filter(published=True).count(), 1)
        self.assertEquals(app_models.Appartment.objects.count(), 2)
        self.assertEquals(app_models.Appartment.objects.dirty().count(), 0)
        self.assertEquals(app_models.Appartment.objects.filter(published_dirty=True, version__isnull=True).count(), 0)
        self.assertEquals(app_models.Appartment.objects.filter(published=True).count(), 1)
        self.assertEquals(app_models.CleaningLady.objects.count(), 4)
        self.assertEquals(app_models.CleaningLady.objects.dirty().count(), 0)
        self.assertEquals(app_models.CleaningLady.objects.filter(published_dirty=True, version__isnull=True).count(), 0)
        self.assertEquals(app_models.CleaningLady.objects.filter(published=True).count(), 2)
        # appartment_cleaninglady is published
        self.assertEquals(app_models.Appartment_CleaningLady.objects.count(), 4)
        self.assertEquals(app_models.Appartment_CleaningLady.objects.dirty().count(), 0)
        self.assertEquals(app_models.Appartment_CleaningLady.objects.filter(published_dirty=True, version__isnull=True).count(), 0)
        self.assertEquals(app_models.Appartment_CleaningLady.objects.filter(published=True).count(), 3)

        # check versions
        a1 = app_models.Appartment.objects.get_versionedobject(name='a1', version=v1)
        self.assertEquals(repr(a1.building), u'<Building: b1 (v1)>')
        self.assertEquals(repr(a1.cleaningladies.all()), u'[<CleaningLady: c1 (v1)>]')
        c1 = app_models.CleaningLady.objects.get_versionedobject(name='c1', version=v1)
        self.assertEquals(repr(c1.appartment_set.all()), u'[<Appartment: a1 (v1)>]')
        c2 = app_models.CleaningLady.objects.get_versionedobject(name='c2', version=v1)
        self.assertEquals(repr(c2.appartment_set.all()), u'[]')

        a1 = app_models.Appartment.objects.get_versionedobject(name='a1', version=v2)
        self.assertEquals(repr(a1.building), u'<Building: b1 (v1)>')
        self.assertEquals(repr(a1.cleaningladies.all()), u'[<CleaningLady: c2 (v1)>]')
        c1 = app_models.CleaningLady.objects.get_versionedobject(name='c1', version=v2)
        self.assertEquals(repr(c1.appartment_set.all()), u'[]')
        c2 = app_models.CleaningLady.objects.get_versionedobject(name='c2', version=v2)
        self.assertEquals(repr(c2.appartment_set.all()), u'[<Appartment: a1 (v1)>]')

        a1 = app_models.Appartment.objects.get_versionedobject(name='a1', version=WorkVersion)
        self.assertEquals(repr(a1.building), u'<Building: b1a>')
        self.assertEquals(repr(a1.cleaningladies.all()), u'[<CleaningLady: c2>]')
        c1 = app_models.CleaningLady.objects.get_versionedobject(name='c1', version=WorkVersion)
        self.assertEquals(repr(c1.appartment_set.all()), u'[]')
        c2 = app_models.CleaningLady.objects.get_versionedobject(name='c2', version=WorkVersion)
        self.assertEquals(repr(c2.appartment_set.all()), u'[<Appartment: a1>]')


class ViewsHistory(TestCase):
    fixtures = ['dummyusers',]

    def setUp(self):
        from django.contrib.auth.models import User, Permission
        from django.contrib.contenttypes.models import ContentType
        # give the users permissions

        app_label = 'versioning'

        # authorisations for users:
        # 1 = authorised
        # 2 = authorised to view all versions
        # 3 = authorised to view public versions
        # 4 = unauthorised

        u = User.objects.get(pk=1)
        u.user_permissions.add(Permission.objects.get(codename='display_published_versions', content_type=ContentType.objects.get(app_label=app_label, model='version')))
        u.user_permissions.add(Permission.objects.get(codename='display_versions', content_type=ContentType.objects.get(app_label=app_label, model='version')))
        u.user_permissions.add(Permission.objects.get(codename='edit_versions', content_type=ContentType.objects.get(app_label=app_label, model='version')))

        u = User.objects.get(pk=2)
        u.user_permissions.add(Permission.objects.get(codename='display_published_versions', content_type=ContentType.objects.get(app_label=app_label, model='version')))
        u.user_permissions.add(Permission.objects.get(codename='display_versions', content_type=ContentType.objects.get(app_label=app_label, model='version')))

        u = User.objects.get(pk=3)
        u.user_permissions.add(Permission.objects.get(codename='display_published_versions', content_type=ContentType.objects.get(app_label=app_label, model='version')))

        # create objects and versions
        user = User.objects.get(pk=1)
        self.assertEquals(app_models.Building.objects.count(), 0)

        b1 = app_models.Building(name='b1')
        b1.save(user)
        v1 = Version(name='v1')
        v1.save(user)

        b1 = app_models.Building.objects.get_versionedobject(v_id = 1, version=WorkVersion)
        b1.name = 'b1a'
        b1.save(user)

        # wait a bit to distinguish the datetimes on the versions
        time.sleep(0.01)
        v2 = Version(name='v2')
        v2.save(user)

        b1 = app_models.Building.objects.get_versionedobject(v_id = 1, version=WorkVersion)
        b1.name = 'b1b'
        b1.save(user)

        # wait a bit to distinguish the datetimes on the versions
        time.sleep(0.01)
        v3 = Version(name='v3', published=True)
        v3.save(user)


    def test_history_views(self):

        defaultuser = User.objects.get(pk=1)
        users = (0,1,2,3,4)

        class history(utils_tests.TestDefinition):
            """
            The history view includes all versions.
            For user 3, only public versions are included.
            """
            description = 'display version 1'
            url = '/versioning/history-for/versioning_test/building/1/'
            users = [1,2,3,]
            contains = [
                {'url': 'v1', 'users': [1,2]},
                {'url': 'v2', 'users': [1,2]},
                {'url': 'v3', 'users': [1,2,3]},
            ]

        tests = [
            history(),
        ]

        utils_tests.run_tests(self, users, tests)

class TestRestoreTestCase(TestCase):
    fixtures = ['dummyusers',]
    def setUp(self):
        from django.contrib.auth.models import User, Permission
        from django.contrib.contenttypes.models import ContentType
        # give the users permissions

        app_label = 'versioning'

        u = User.objects.get(pk=1)
        u.user_permissions.add(Permission.objects.get(codename='display_published_versions', content_type=ContentType.objects.get(app_label=app_label, model='version')))
        u.user_permissions.add(Permission.objects.get(codename='display_versions', content_type=ContentType.objects.get(app_label=app_label, model='version')))
        u.user_permissions.add(Permission.objects.get(codename='edit_versions', content_type=ContentType.objects.get(app_label=app_label, model='version')))

class RestoreSimple(TestRestoreTestCase):
    def test_restore_changed(self):
        user = User.objects.get(pk=1)

        self.assertEquals(app_models.Building.objects.count(), 0)
        self.assertEquals(app_models.Building.objects.next_v_id(), 1)

        b1 = app_models.Building(name='b1')
        b1.save(user)

        v1 = Version(name='v1')
        v1.save(user)

        b1 = app_models.Building.objects.get_versionedobject(v_id = 1, version=WorkVersion)
        b1.name = 'b1a'
        b1.save(user)

        # wait a bit to distinguish the datetimes on the versions
        time.sleep(0.01)
        v2 = Version(name='v2')
        v2.save(user)

        b = app_models.Building.objects.get_versionedobject(v_id=1, version=WorkVersion)
        self.assertEquals(b.name, 'b1a')

        b.restore_version(v1)

        b = app_models.Building.objects.get_versionedobject(v_id=1, version=WorkVersion)
        self.assertEquals(b.name, 'b1')

    def test_restore_new(self):
        user = User.objects.get(pk=1)

        self.assertEquals(app_models.Building.objects.count(), 0)
        self.assertEquals(app_models.Building.objects.next_v_id(), 1)

        v1 = Version(name='v1')
        v1.save(user)

        b1 = app_models.Building(name='b1')
        b1.save(user)

        # wait a bit to distinguish the datetimes on the versions
        time.sleep(0.01)
        v2 = Version(name='v2')
        v2.save(user)

        b = app_models.Building.objects.get_versionedobject(v_id=1, version=WorkVersion)
        self.assertEquals(b.name, 'b1')

        b.restore_version(v1)

        b = app_models.Building.objects.get(v_id=1, version__isnull=True)
        self.assertEquals(b.deleted, True)

    def test_get_newer_objects(self):
        """
            Version 1     b1   b2   b3
            Version 2          b2a      b4
            Version 3     b1-
        """
        user = User.objects.get(pk=1)

        b1 = app_models.Building(name='b1')
        b1.save(user)

        b2 = app_models.Building(name='b2')
        b2.save(user)

        b3 = app_models.Building(name='b3')
        b3.save(user)

        v1 = Version(name='v1')
        v1.save(user)

        b2 = app_models.Building.objects.get_versionedobject(v_id=2, version=WorkVersion)
        b2.name = 'b2a'
        b2.save(user)

        b4 = app_models.Building(name='b4')
        b4.save(user)

        # wait a bit to distinguish the datetimes on the versions
        time.sleep(0.01)
        v2 = Version(name='v2')
        v2.save(user)

        b1 = app_models.Building.objects.get_versionedobject(v_id=1, version=WorkVersion)
        b1.delete(user)

        # wait a bit to distinguish the datetimes on the versions
        time.sleep(0.01)
        v3 = Version(name='v3')
        v3.save(user)

        # This should be empty
        qs = app_models.Building.objects.get_newerobjects(version=WorkVersion)
        self.assertEquals(len(qs), 0)

        # This should be empty
        qs = app_models.Building.objects.get_newerobjects(version=None)
        self.assertEquals(len(qs), 0)

        # This should include b1-, b2a, b4
        qs = app_models.Building.objects.order_by('name').get_newerobjects(version=v1)
        self.assertEquals(len(qs), 3)
        self.assertEquals(qs[0].name, 'b1')
        self.assertEquals(qs[0].deleted, True)
        self.assertEquals(qs[1].name, 'b2a')
        self.assertEquals(qs[2].name, 'b4')

        # This should include b1-
        qs = app_models.Building.objects.get_newerobjects(version=v2)
        self.assertEquals(len(qs), 1)
        self.assertEquals(qs[0].name, 'b1')
        self.assertEquals(qs[0].deleted, True)

        # This should be empty
        qs = app_models.Building.objects.get_newerobjects(version=v3)
        self.assertEquals(len(qs), 0)

class RestoreFK(TestRestoreTestCase):
    def test_restore_fk(self):
        user = User.objects.get(pk=1)

        self.assertEquals(app_models.Building.objects.count(), 0)
        self.assertEquals(app_models.Appartment.objects.count(), 0)
        self.assertEquals(app_models.Building.objects.next_v_id(), 1)
        self.assertEquals(app_models.Appartment.objects.next_v_id(), 1)

        b1 = app_models.Building(name='b1')
        b1.save(user)

        a1 = app_models.Appartment(name='a1', building=b1)
        a1.save(user)

        v1 = Version(name='v1')
        v1.save(user)

        a1 = app_models.Appartment.objects.get_versionedobject(v_id = 1, version=WorkVersion)
        a1.name = 'a1a'
        a1.save(user)

        # wait a bit to distinguish the datetimes on the versions
        time.sleep(0.01)
        v2 = Version(name='v2')
        v2.save(user)

        a = app_models.Appartment.objects.get_versionedobject(v_id = 1, version=WorkVersion)
        self.assertEquals(a.name, 'a1a')

        a.restore_version(v1)

        a = app_models.Appartment.objects.get_versionedobject(v_id=1, version=WorkVersion)
        self.assertEquals(a.name, 'a1')

        # now delete the building

        b1.delete(user)

        # This should raise an exception
        self.assertRaises(Exception, a.restore_version, v1)

        # restoring b first should work
        b = app_models.Building.objects.get_versionedobject(v_id=1, version=v1)
        b.restore_version(v1)
        a = app_models.Appartment.objects.get_versionedobject(v_id=1, version=WorkVersion)
        a.restore_version(v1)

class RestoreObjects(TestRestoreTestCase):
    def test_restore_objects(self):
        user = User.objects.get(pk=1)

        self.assertEquals(app_models.Building.objects.count(), 0)
        self.assertEquals(app_models.Appartment.objects.count(), 0)
        self.assertEquals(app_models.Building.objects.next_v_id(), 1)
        self.assertEquals(app_models.Appartment.objects.next_v_id(), 1)

        b1 = app_models.Building(name='b1')
        b1.save(user)

        a1 = app_models.Appartment(name='a1', building=b1)
        a1.save(user)

        v1 = Version(name='v1')
        v1.save(user)

        a1 = app_models.Appartment.objects.get_versionedobject(v_id = 1, version=WorkVersion)
        a1.name = 'a1a'
        a1.save(user)

        b1 = app_models.Building.objects.get_versionedobject(v_id=1, version=WorkVersion)
        b1.name = 'b1a'
        b1.save(user)

        # wait a bit to distinguish the datetimes on the versions
        time.sleep(0.01)
        v2 = Version(name='v2')
        v2.save(user)

        a = app_models.Appartment.objects.get_versionedobject(v_id = 1, version=WorkVersion)
        self.assertEquals(a.name, 'a1a')
        b = app_models.Building.objects.get_versionedobject(v_id=1, version=WorkVersion)
        self.assertEquals(b.name, 'b1a')

        v1.restore_objects(user, [[b],[a]])

        a = app_models.Appartment.objects.get_versionedobject(v_id = 1, version=WorkVersion)
        self.assertEquals(a.name, 'a1')
        b = app_models.Building.objects.get_versionedobject(v_id=1, version=WorkVersion)
        self.assertEquals(b.name, 'b1')

        v2.restore_objects(user, [[b],[a]])

        a = app_models.Appartment.objects.get_versionedobject(v_id = 1, version=WorkVersion)
        self.assertEquals(a.name, 'a1a')
        b = app_models.Building.objects.get_versionedobject(v_id=1, version=WorkVersion)
        self.assertEquals(b.name, 'b1a')

class RestoreVersion(TestRestoreTestCase):
    def test_restore_version(self):
        user = User.objects.get(pk=1)

        self.assertEquals(app_models.Building.objects.count(), 0)
        self.assertEquals(app_models.Appartment.objects.count(), 0)
        self.assertEquals(app_models.Building.objects.next_v_id(), 1)
        self.assertEquals(app_models.Appartment.objects.next_v_id(), 1)

        b1 = app_models.Building(name='b1')
        b1.save(user)

        a1 = app_models.Appartment(name='a1', building=b1)
        a1.save(user)

        v1 = Version(name='v1')
        v1.save(user)

        a1 = app_models.Appartment.objects.get_versionedobject(v_id = 1, version=WorkVersion)
        a1.name = 'a1a'
        a1.save(user)

        b1 = app_models.Building.objects.get_versionedobject(v_id=1, version=WorkVersion)
        b1.name = 'b1a'
        b1.save(user)

        # wait a bit to distinguish the datetimes on the versions
        time.sleep(0.01)
        v2 = Version(name='v2')
        v2.save(user)

        a = app_models.Appartment.objects.get_versionedobject(v_id = 1, version=WorkVersion)
        self.assertEquals(a.name, 'a1a')
        b = app_models.Building.objects.get_versionedobject(v_id=1, version=WorkVersion)
        self.assertEquals(b.name, 'b1a')

        v1.restore(user)

        a = app_models.Appartment.objects.get_versionedobject(v_id = 1, version=WorkVersion)
        self.assertEquals(a.name, 'a1')
        b = app_models.Building.objects.get_versionedobject(v_id=1, version=WorkVersion)
        self.assertEquals(b.name, 'b1')

        v2.restore(user)

        a = app_models.Appartment.objects.get_versionedobject(v_id = 1, version=WorkVersion)
        self.assertEquals(a.name, 'a1a')
        b = app_models.Building.objects.get_versionedobject(v_id=1, version=WorkVersion)
        self.assertEquals(b.name, 'b1a')

class RestoreTree(TestRestoreTestCase):
    def test_treemodels(self):
        user = User.objects.get(pk=1)

        self.assertEquals(app_models.DiskTreeElement.objects.count(), 0)
        self.assertEquals(app_models.Folder.objects.count(), 0)

        # create a tree element
        folder = app_models.Folder(name=u'Base')
        treeelement = folder.save(user, parent=None, treeelementkwargs={})
        self.assertEquals(folder.tree_display(), u'Base')
        self.assertEquals(folder.get_elementtype_for_tree(), u'event')
        self.assertEquals(folder.get_treeelements().count(), 1)

        # and another
        folder = app_models.Folder(name=u'Child 1')
        treeelementc1 = folder.save(user, parent=app_models.DiskTreeElement.objects.get(pk=treeelement.pk), treeelementkwargs={})

        # and another
        folder = app_models.Folder(name=u'Child 2')
        treeelementc2 = folder.save(user, parent=app_models.DiskTreeElement.objects.get(pk=treeelement.pk), treeelementkwargs={})

        tree = app_models.DiskTreeElement.tree.get_versionedobjects(version=WorkVersion)
        self.assertEquals(print_tree(tree), '-Base|--Child 1|--Child 2')

        v1 = Version(name='v1')
        v1.save(user)

        # test modifying a leaf
        folder = app_models.Folder.objects.get_versionedobject(name='Child 1', version=WorkVersion)
        folder.name = u'Modified child 1'
        folder.save(user)

        tree = app_models.DiskTreeElement.tree.get_versionedobjects(version=WorkVersion)
        self.assertEquals(print_tree(tree), '-Base|--Modified child 1|--Child 2')

        # wait a bit to distinguish the datetimes on the versions
        time.sleep(0.01)
        v2 = Version(name='v2')
        v2.save(user)

        v1.restore(user)

        tree = app_models.DiskTreeElement.tree.get_versionedobjects(version=WorkVersion)
        self.assertEquals(print_tree(tree), '-Base|--Child 1|--Child 2')

        # now test modifying the tree
        # move child 2 underneath child 1
        self.assertEquals(app_models.DiskTreeElement.objects.get_versionedobject(v_id=3, version=WorkVersion).parent.v_id, 1)
        el1 = app_models.DiskTreeElement.objects.get_versionedobject(v_id=2, version=WorkVersion)
        el2 = app_models.DiskTreeElement.objects.get_versionedobject(v_id=3, version=WorkVersion)
        el2.move_to(user, el1)

        tree = app_models.DiskTreeElement.tree.get_versionedobjects(version=WorkVersion)
        self.assertEquals(print_tree(tree), '-Base|--Child 1|---Child 2')

        # wait a bit to distinguish the datetimes on the versions
        time.sleep(0.01)
        v3 = Version(name='v3')
        v3.save(user)

        # add another folder
        folder = app_models.Folder.objects.get_versionedobject(name='Child 1', version=WorkVersion)
        treeelementc1 = folder.get_treeelements()[0]

        folder = app_models.Folder(name=u'Child 3')
        folder.save(user, parent=app_models.DiskTreeElement.objects.get(pk=treeelementc1.pk), treeelementkwargs={})

        tree = app_models.DiskTreeElement.tree.get_versionedobjects(version=WorkVersion)
        self.assertEquals(print_tree(tree), '-Base|--Child 1|---Child 2|---Child 3')

        v1.restore(user)

        tree = app_models.DiskTreeElement.tree.get_versionedobjects(version=WorkVersion)
        self.assertEquals(print_tree(tree), '-Base|--Child 1|--Child 2')

        # now test with deleting elements

        v3.restore(user)

        tree = app_models.DiskTreeElement.tree.get_versionedobjects(version=WorkVersion)
        self.assertEquals(print_tree(tree), '-Base|--Child 1|---Child 2')

        # delete children
        folder = app_models.Folder.objects.get_versionedobject(name='Child 2', version=WorkVersion)
        folder.delete(user)

        folder = app_models.Folder.objects.get_versionedobject(name='Child 1', version=WorkVersion)
        folder.delete(user)

        tree = app_models.DiskTreeElement.tree.get_versionedobjects(version=WorkVersion)
        self.assertEquals(print_tree(tree), '-Base')

        # wait a bit to distinguish the datetimes on the versions
        time.sleep(0.01)
        v4 = Version(name='v4')
        v4.save(user)

        v3.restore(user)

        tree = app_models.DiskTreeElement.tree.get_versionedobjects(version=WorkVersion)
        self.assertEquals(print_tree(tree), '-Base|--Child 1|---Child 2')

        time.sleep(0.01)
        v5 = Version(name='v5')
        v5.save(user)

        # delete the original root
        # add a new base
        folder = app_models.Folder(name=u'Newbase')
        treeelementnb = folder.save(user, parent=None, treeelementkwargs={})
        treeelementnb.move_to(user, treeelement, 'right')
        folder = app_models.Folder.objects.get_versionedobject(name='Base', version=WorkVersion)
        treeelement = folder.get_treeelements()[0]
        #app_models.DiskTreeElement.objects.get(id=treeelement.id)
        treeelementnb = app_models.DiskTreeElement.objects.get(id=treeelementnb.id)
        treeelement.move_to(user, treeelementnb, 'last-child')

        tree = app_models.DiskTreeElement.tree.get_versionedobjects(version=WorkVersion)
        self.assertEquals(print_tree(tree), '-Newbase|--Base|---Child 1|----Child 2')

        # delete children
        folder = app_models.Folder.objects.get_versionedobject(name='Child 2', version=WorkVersion)
        folder.delete(user)

        folder = app_models.Folder.objects.get_versionedobject(name='Child 1', version=WorkVersion)
        folder.delete(user)

        folder = app_models.Folder.objects.get_versionedobject(name='Base', version=WorkVersion)
        folder.delete(user)

        tree = app_models.DiskTreeElement.tree.get_versionedobjects(version=WorkVersion)
        self.assertEquals(print_tree(tree), '-Newbase')

        # wait a bit to distinguish the datetimes on the versions
        time.sleep(0.01)
        v6 = Version(name='v6')
        v6.save(user)

        v5.restore(user)
        tree = app_models.DiskTreeElement.tree.get_versionedobjects(version=WorkVersion)
        self.assertEquals(print_tree(tree), '-Base|--Child 1|---Child 2')

        # wait a bit to distinguish the datetimes on the versions
        time.sleep(0.01)
        v7 = Version(name='v7')
        v7.save(user)

        v3.restore(user)
        tree = app_models.DiskTreeElement.tree.get_versionedobjects(version=WorkVersion)
        self.assertEquals(print_tree(tree), '-Base|--Child 1|---Child 2')


        tree = app_models.DiskTreeElement.tree.get_versionedobjects(version=v1)
        self.assertEquals(print_tree(tree), '-Base (v1)|--Child 1 (v1)|--Child 2 (v1)')

        tree = app_models.DiskTreeElement.tree.get_versionedobjects(version=v2)
        self.assertEquals(print_tree(tree), '-Base (v1)|--Modified child 1 (v1)|--Child 2 (v1)')

        tree = app_models.DiskTreeElement.tree.get_versionedobjects(version=v3)
        self.assertEquals(print_tree(tree), '-Base (v3)|--Child 1 (v3)|---Child 2 (v3)')

##        for item in app_models.DiskTreeElement.objects.all():
##            print item.id, item.v_id, item.parent_id, item, item.version, item.tree_id, item.deleted

        tree = app_models.DiskTreeElement.tree.get_versionedobjects(version=v4)
        self.assertEquals(print_tree(tree), '-Base (v4)')

        tree = app_models.DiskTreeElement.tree.get_versionedobjects(version=v5)
        self.assertEquals(print_tree(tree), '-Base (v5)|--Child 1 (v5)|---Child 2 (v5)')

        tree = app_models.DiskTreeElement.tree.get_versionedobjects(version=v6)
        self.assertEquals(print_tree(tree), '-Newbase (v6)')

        tree = app_models.DiskTreeElement.tree.get_versionedobjects(version=v7)
        self.assertEquals(print_tree(tree), '-Base (v7)|--Child 1 (v7)|---Child 2 (v7)')

    def test_2(self):
        user = User.objects.get(pk=1)

        self.assertEquals(app_models.DiskTreeElement.objects.count(), 0)
        self.assertEquals(app_models.Folder.objects.count(), 0)

        # create some trees
        folder = app_models.Folder(name=u'A1')
        treeelementa1 = folder.save(user, parent=None, treeelementkwargs={})
        folder = app_models.Folder(name=u'A1B1')
        treeelementa1b1 = folder.save(user, parent=app_models.DiskTreeElement.objects.get(pk=treeelementa1.pk), treeelementkwargs={})
        folder = app_models.Folder(name=u'A1B1C1')
        treeelementa1b1c1 = folder.save(user, parent=app_models.DiskTreeElement.objects.get(pk=treeelementa1b1.pk), treeelementkwargs={})
        folder = app_models.Folder(name=u'A2')
        treeelementa2 = folder.save(user, parent=None, treeelementkwargs={})
        folder = app_models.Folder(name=u'A2B1')
        treeelementa2b1 = folder.save(user, parent=app_models.DiskTreeElement.objects.get(pk=treeelementa2.pk), treeelementkwargs={})
        folder = app_models.Folder(name=u'A2B1C1')
        treeelementa2b1c1 = folder.save(user, parent=app_models.DiskTreeElement.objects.get(pk=treeelementa2b1.pk), treeelementkwargs={})
        folder = app_models.Folder(name=u'A3')
        treeelementa3 = folder.save(user, parent=None, treeelementkwargs={})
        folder = app_models.Folder(name=u'A3B1')
        treeelementa3b1 = folder.save(user, parent=app_models.DiskTreeElement.objects.get(pk=treeelementa3.pk), treeelementkwargs={})
        folder = app_models.Folder(name=u'A3B1C1')
        treeelementa3b1c1 = folder.save(user, parent=app_models.DiskTreeElement.objects.get(pk=treeelementa3b1.pk), treeelementkwargs={})

        tree = app_models.DiskTreeElement.tree.get_versionedobjects(version=WorkVersion)
        self.assertEquals(print_tree(tree), '-A1|--A1B1|---A1B1C1|-A2|--A2B1|---A2B1C1|-A3|--A3B1|---A3B1C1')

        v1 = Version(name='v1')
        v1.save(user)

        tree = app_models.DiskTreeElement.tree.get_versionedobjects(version=WorkVersion)
        self.assertEquals(print_tree(tree), '-A1|--A1B1|---A1B1C1|-A2|--A2B1|---A2B1C1|-A3|--A3B1|---A3B1C1')

        # edit a tree
        # add a new base
        folder = app_models.Folder(name=u'C1')
        treeelementc1 = folder.save(user, parent=None, treeelementkwargs={})
        treeelementc1.move_to(user, app_models.DiskTreeElement.objects.get(pk=treeelementa1.pk), 'right')
        treeelementa1 = app_models.DiskTreeElement.objects.get(id=treeelementa1.pk)
        treeelementa1.move_to(user, treeelementc1, 'last-child')
        folder = app_models.Folder.objects.get_versionedobject(name='A1B1C1', version=WorkVersion)
        folder.delete(user)

        # delete a tree
        folder = app_models.Folder.objects.get_versionedobject(name='A2B1C1', version=WorkVersion)
        folder.delete(user)
        folder = app_models.Folder.objects.get_versionedobject(name='A2B1', version=WorkVersion)
        folder.delete(user)
        folder = app_models.Folder.objects.get_versionedobject(name='A2', version=WorkVersion)
        folder.delete(user)

        # add a tree
        folder = app_models.Folder(name=u'A4')
        treeelementa4 = folder.save(user, parent=None, treeelementkwargs={})
        folder = app_models.Folder(name=u'A4B1')
        treeelementa4b1 = folder.save(user, parent=app_models.DiskTreeElement.objects.get(pk=treeelementa4.pk), treeelementkwargs={})
        folder = app_models.Folder(name=u'A4B1C1')
        treeelementa4b1c1 = folder.save(user, parent=app_models.DiskTreeElement.objects.get(pk=treeelementa4b1.pk), treeelementkwargs={})

        tree = app_models.DiskTreeElement.tree.get_versionedobjects(version=WorkVersion)
        self.assertEquals(print_tree(tree), '-C1|--A1|---A1B1|-A3|--A3B1|---A3B1C1|-A4|--A4B1|---A4B1C1')

        v2 = Version(name='v2')
        v2.save(user)

        tree = app_models.DiskTreeElement.tree.get_versionedobjects(version=WorkVersion)
        self.assertEquals(print_tree(tree), '-C1|--A1|---A1B1|-A3|--A3B1|---A3B1C1|-A4|--A4B1|---A4B1C1')

        tree = app_models.DiskTreeElement.tree.get_versionedobjects(version=v1)
        self.assertEquals(print_tree(tree), '-A1 (v1)|--A1B1 (v1)|---A1B1C1 (v1)|-A2 (v1)|--A2B1 (v1)|---A2B1C1 (v1)|-A3 (v1)|--A3B1 (v1)|---A3B1C1 (v1)')

    # ! this means order of separate trees is not preserved when versioning !!!
        tree = app_models.DiskTreeElement.tree.get_versionedobjects(version=v2)
        self.assertEquals(print_tree(tree), '-A3 (v1)|--A3B1 (v1)|---A3B1C1 (v1)|-C1 (v2)|--A1 (v2)|---A1B1 (v2)|-A4 (v2)|--A4B1 (v2)|---A4B1C1 (v2)')

        v1.restore(user)

        tree = app_models.DiskTreeElement.tree.get_versionedobjects(version=WorkVersion)
        self.assertEquals(print_tree(tree), '-A1|--A1B1|---A1B1C1|-A2|--A2B1|---A2B1C1|-A3|--A3B1|---A3B1C1')

        tree = app_models.DiskTreeElement.tree.get_versionedobjects(version=v1)
        self.assertEquals(print_tree(tree), '-A1 (v1)|--A1B1 (v1)|---A1B1C1 (v1)|-A2 (v1)|--A2B1 (v1)|---A2B1C1 (v1)|-A3 (v1)|--A3B1 (v1)|---A3B1C1 (v1)')

        tree = app_models.DiskTreeElement.tree.get_versionedobjects(version=v2)
        self.assertEquals(print_tree(tree), '-A3 (v1)|--A3B1 (v1)|---A3B1C1 (v1)|-C1 (v2)|--A1 (v2)|---A1B1 (v2)|-A4 (v2)|--A4B1 (v2)|---A4B1C1 (v2)')

class MergeVersion(TestRestoreTestCase):
    def test_merge_simple(self):
        user = User.objects.get(pk=1)

        self.assertEquals(app_models.Building.objects.count(), 0)
        self.assertEquals(app_models.Appartment.objects.count(), 0)
        self.assertEquals(app_models.Building.objects.next_v_id(), 1)
        self.assertEquals(app_models.Appartment.objects.next_v_id(), 1)

        b1 = app_models.Building(name='b1')
        b1.save(user)

        a1 = app_models.Appartment(name='a1', building=b1)
        a1.save(user)

        v1 = Version(name='v1')
        v1.save(user)

        a1 = app_models.Appartment.objects.get_versionedobject(v_id = 1, version=WorkVersion)
        a1.name = 'a1a'
        a1.save(user)

        # wait a bit to distinguish the datetimes on the versions
        time.sleep(0.01)
        v2 = Version(name='v2')
        v2.save(user)

        self.assertEquals(Version.objects.count(), 2)
        self.assertEquals(app_models.Building.objects.count(), 2)
        self.assertEquals(app_models.Appartment.objects.count(), 3)

        a = app_models.Appartment.objects.get_versionedobject(v_id = 1, version=v2)
        self.assertEquals(a.name, 'a1a')
        a = app_models.Appartment.objects.get_versionedobject(v_id = 1, version=v1)
        self.assertEquals(a.name, 'a1')
        b = app_models.Building.objects.get_versionedobject(v_id=1, version=v2)
        self.assertEquals(b.name, 'b1')
        b = app_models.Building.objects.get_versionedobject(v_id=1, version=v1)
        self.assertEquals(b.name, 'b1')

        v2.merge_previous(user)

        self.assertEquals(Version.objects.count(), 1)
        self.assertEquals(app_models.Building.objects.count(), 2)
        self.assertEquals(app_models.Appartment.objects.count(), 2)

        a = app_models.Appartment.objects.get_versionedobject(v_id = 1, version=v2)
        self.assertEquals(a.name, 'a1a')
        b = app_models.Building.objects.get_versionedobject(v_id=1, version=v2)
        self.assertEquals(b.name, 'b1')
        self.assertEquals(a.building, b)

    def test_merge_treemodels(self):
        user = User.objects.get(pk=1)

        self.assertEquals(app_models.DiskTreeElement.objects.count(), 0)
        self.assertEquals(app_models.Folder.objects.count(), 0)

        # create a tree element
        folder = app_models.Folder(name=u'Base')
        treeelement = folder.save(user, parent=None, treeelementkwargs={})

        # and another
        folder = app_models.Folder(name=u'Child 1')
        treeelementc1 = folder.save(user, parent=app_models.DiskTreeElement.objects.get(pk=treeelement.pk), treeelementkwargs={})

        # and another
        folder = app_models.Folder(name=u'Child 2')
        treeelementc2 = folder.save(user, parent=app_models.DiskTreeElement.objects.get(pk=treeelement.pk), treeelementkwargs={})

        tree = app_models.DiskTreeElement.tree.get_versionedobjects(version=WorkVersion)
        self.assertEquals(print_tree(tree), '-Base|--Child 1|--Child 2')

        self.assertEquals(app_models.DiskTreeElement.objects.count(), 3)
        self.assertEquals(app_models.Folder.objects.count(), 3)

        v1 = Version(name='v1')
        v1.save(user)

        self.assertEquals(app_models.DiskTreeElement.objects.count(), 6)
        self.assertEquals(app_models.Folder.objects.count(), 6)

        # test modifying a leaf
        folder = app_models.Folder.objects.get_versionedobject(name='Child 1', version=WorkVersion)
        folder.name = u'Modified child 1'
        folder.save(user)

        # wait a bit to distinguish the datetimes on the versions
        time.sleep(0.01)
        v2 = Version(name='v2')
        v2.save(user)

        self.assertEquals(app_models.DiskTreeElement.objects.count(), 6)
        self.assertEquals(app_models.Folder.objects.count(), 7)

        tree = app_models.DiskTreeElement.tree.get_versionedobjects(version=WorkVersion)
        self.assertEquals(print_tree(tree), '-Base|--Modified child 1|--Child 2')

        # move child 2 underneath child 1
        self.assertEquals(app_models.DiskTreeElement.objects.get_versionedobject(v_id=3, version=WorkVersion).parent.v_id, 1)
        el1 = app_models.DiskTreeElement.objects.get_versionedobject(v_id=2, version=WorkVersion)
        el2 = app_models.DiskTreeElement.objects.get_versionedobject(v_id=3, version=WorkVersion)
        el2.move_to(user, el1)

        tree = app_models.DiskTreeElement.tree.get_versionedobjects(version=WorkVersion)
        self.assertEquals(print_tree(tree), '-Base|--Modified child 1|---Child 2')

        # wait a bit to distinguish the datetimes on the versions
        time.sleep(0.01)
        v3 = Version(name='v3')
        v3.save(user)

        self.assertEquals(app_models.DiskTreeElement.objects.count(), 9)
        self.assertEquals(app_models.Folder.objects.count(), 7)

        # add another folder
        folder = app_models.Folder(name=u'Child 3')
        folder.save(user, parent=app_models.DiskTreeElement.objects.get(pk=treeelementc1.pk), treeelementkwargs={})

        tree = app_models.DiskTreeElement.tree.get_versionedobjects(version=WorkVersion)
        self.assertEquals(print_tree(tree), '-Base|--Modified child 1|---Child 2|---Child 3')

        # wait a bit to distinguish the datetimes on the versions
        time.sleep(0.01)
        v4 = Version(name='v4')
        v4.save(user)

        self.assertEquals(app_models.DiskTreeElement.objects.count(), 14)
        self.assertEquals(app_models.Folder.objects.count(), 9)

        # Let's make the workversion different
        folder = app_models.Folder.objects.get_versionedobject(name='Child 2', version=WorkVersion)
        folder.name = u'Modified child 2'
        folder.save(user)

        # Let's look at the versions:

        tree = app_models.DiskTreeElement.tree.get_versionedobjects(version=v1)
        self.assertEquals(print_tree(tree), '-Base (v1)|--Child 1 (v1)|--Child 2 (v1)')

        tree = app_models.DiskTreeElement.tree.get_versionedobjects(version=v2)
        self.assertEquals(print_tree(tree), '-Base (v1)|--Modified child 1 (v1)|--Child 2 (v1)')

        tree = app_models.DiskTreeElement.tree.get_versionedobjects(version=v3)
        self.assertEquals(print_tree(tree), '-Base (v3)|--Modified child 1 (v3)|---Child 2 (v3)')

        tree = app_models.DiskTreeElement.tree.get_versionedobjects(version=v4)
        self.assertEquals(print_tree(tree), '-Base (v4)|--Modified child 1 (v4)|---Child 2 (v4)|---Child 3 (v4)')

        tree = app_models.DiskTreeElement.tree.get_versionedobjects(version=WorkVersion)
        self.assertEquals(print_tree(tree), '-Base|--Modified child 1|---Modified child 2|---Child 3')

        self.assertEquals(Version.objects.count(), 4)

        # Let's merge

        # this removes v3
        v4.merge_previous(user)
        self.assertEquals(app_models.DiskTreeElement.objects.count(), 11)
        self.assertEquals(app_models.Folder.objects.count(), 9)

        tree = app_models.DiskTreeElement.tree.get_versionedobjects(version=v4)
        self.assertEquals(print_tree(tree), '-Base (v4)|--Modified child 1 (v4)|---Child 2 (v4)|---Child 3 (v4)')

        self.assertEquals(Version.objects.count(), 3)

        # this removes v2
        v4.merge_previous(user)
        self.assertEquals(app_models.DiskTreeElement.objects.count(), 11)
        self.assertEquals(app_models.Folder.objects.count(), 9)

        tree = app_models.DiskTreeElement.tree.get_versionedobjects(version=v4)
        self.assertEquals(print_tree(tree), '-Base (v4)|--Modified child 1 (v4)|---Child 2 (v4)|---Child 3 (v4)')

        self.assertEquals(Version.objects.count(), 2)

        # this removes v1
        v4.merge_previous(user)
        self.assertEquals(app_models.DiskTreeElement.objects.count(), 8)
        self.assertEquals(app_models.Folder.objects.count(), 8)

        tree = app_models.DiskTreeElement.tree.get_versionedobjects(version=v4)
        self.assertEquals(print_tree(tree), '-Base (v4)|--Modified child 1 (v4)|---Child 2 (v4)|---Child 3 (v4)')

        self.assertEquals(Version.objects.count(), 1)

        # Verify the workversion
        tree = app_models.DiskTreeElement.tree.get_versionedobjects(version=WorkVersion)
        self.assertEquals(print_tree(tree), '-Base|--Modified child 1|---Modified child 2|---Child 3')

class Formfields(TestCase):
    fixtures = ['dummyusers',]
    def test_fkfield(self):
        """
        Create some objects with different id and v_id
        Test correct items used formfield for VersionedFK
        """
        user = User.objects.get(pk=1)

        # build some data

        b1 = app_models.Building(name='b1')
        b1.save(user)

        v1 = Version(name='v1')
        v1.save(user)

        b2 = app_models.Building(name='b2')
        b2.save(user)

        v2 = Version(name='v2')
        v2.save(user)

        class AppartmentForm(VersionedForm):
            class Meta:
                model = app_models.Appartment
        af = AppartmentForm({'name': 'app', 'building': 2})
        self.assertHTMLEqual(str(af['building']), """<select name="building" id="id_building">
            <option value="">---------</option>
            <option value="1">b1</option>
            <option value="2" selected="selected">b2</option>
            </select>""")

